/*-------------------------------------------------------------------------
 *
 * telnet_srv.c
 *		Example of a loadable wire protocol extension using telnet.
 *		
 * Copyright (c) 2021, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *	  contrib/telnet_srv/telnet_srv.c
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include <fcntl.h>
#include <grp.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>

#include "miscadmin.h"
#include "pgstat.h"
#include "access/xact.h"
#include "commands/async.h"
#include "common/ip.h"
#include "common/string.h"
#include "executor/spi.h"
#include "jit/jit.h"
#include "lib/ilist.h"
#include "libpq/crypt.h"
#include "postmaster/interrupt.h"
#include "postmaster/postmaster.h"
#include "postmaster/protocol_extension.h"
#include "storage/ipc.h"
#include "tcop/tcopprot.h"
#include "utils/guc.h"
#include "utils/elog.h"
#include "utils/lsyscache.h"
#include "utils/memutils.h"
#include "utils/pidfile.h"
#include "utils/portal.h"
#include "utils/probes.h"
#include "utils/ps_status.h"
#include "utils/snapmgr.h"
#include "utils/timeout.h"
#include "utils/varlena.h"

PG_MODULE_MAGIC;

/*
 * Configuration options
 */
static int		pe_port;
static char	   *pe_listen_addrs;
static char	   *pe_unix_socket_directories;
static int		pe_unix_socket_permissions;
static char	   *pe_unix_socket_group;

/* Where the Unix socket files are (list of palloc'd strings) */
static List *sock_paths = NIL;

/*---- Function declarations ----*/

void		_PG_init(void);
void		_PG_fini(void);

static void	pe_listen_init(void);
static int	pe_accept(pgsocket server_fd, Port *port);
static void	pe_close(pgsocket server_fd);
static int	pe_start(Port *port);
static void	pe_authenticate(Port *port);
static void	pe_mainfunc(Port *port,
						   int argc, char *argv[]) pg_attribute_noreturn();
static void pe_send_message(ErrorData *edata);
static void	pe_send_ready_for_query(CommandDest dest);
static int	pe_read_command(StringInfo inBuf);
static void	pe_end_command(const QueryCompletion *qc, CommandDest dest,
						   bool force_undecorated_output);
static bool	pe_printtup(TupleTableSlot *slot, DestReceiver *dest);
static void	pe_printtup_startup(DestReceiver *dest, int operation,
								TupleDesc typeinfo);
static void	pe_printtup_shutdown(DestReceiver *dest);
static void	pe_printtup_destroy(DestReceiver *dest);

/* ----------------
 *		Private state for a printtup destination object
 *
 * NOTE: finfo is the lookup info for either typoutput or typsend, whichever
 * we are using for this column.
 * ----------------
 */
typedef struct
{								/* Per-attribute information */
	Oid			typoutput;		/* Oid for the type's text output fn */
	Oid			typsend;		/* Oid for the type's binary output fn */
	bool		typisvarlena;	/* is it varlena (ie possibly toastable)? */
	int16		format;			/* format code for this column */
	FmgrInfo	finfo;			/* Precomputed call info for output fn */
} PrinttupAttrInfo;

typedef struct
{
	DestReceiver pub;			/* publicly-known function pointers */
	Portal		portal;			/* the Portal we are printing from */
	bool		sendDescrip;	/* send RowDescription at startup? */
	TupleDesc	attrinfo;		/* The attr info we are set up for */
	int			nattrs;
	PrinttupAttrInfo *myinfo;	/* Cached info about each attr */
	StringInfoData buf;			/* output buffer (*not* in tmpcontext) */
	MemoryContext tmpcontext;	/* Memory context for per-row workspace */
} DR_printtup;

/*---- Local variables ----*/

static listen_init_hook_type	prev_listen_init;

static ProtocolExtensionConfig pe_config = {
	pe_accept,
	pe_close,
	pe_start,
	pe_authenticate,
	pe_mainfunc,
	pe_send_message,
	NULL,						/* not interested in cancel key */
	pe_send_ready_for_query,
	pe_read_command,
	pe_end_command,
	pe_printtup,
	pe_printtup_startup,
	pe_printtup_shutdown,
	pe_printtup_destroy
};

/*
 * The generic socket support is kept in a separate file
 */
#include "support_funcs.c"


/*
 * Module load callback
 */
void
_PG_init(void)
{
	/*
	 * Define (or redefine) custom GUC variables.
	 */
	DefineCustomIntVariable("telnet_srv.port",
							"Telnet server port.",
							NULL,
							&pe_port,
							pe_port,
							1024,
							65536,
							PGC_POSTMASTER,
							0,
							NULL,
							NULL,
							NULL);

	DefineCustomStringVariable("telnet_srv.listen_addresses",
							   "Telnet server listen addresses.",
							   NULL,
							   &pe_listen_addrs,
							   NULL,
							   PGC_POSTMASTER,
							   0,
							   NULL,
							   NULL,
							   NULL);

	DefineCustomStringVariable("telnet_srv.unix_socket_directories",
							   "Telnet server unix socket directories.",
							   NULL,
							   &pe_unix_socket_directories,
							   NULL,
							   PGC_POSTMASTER,
							   0,
							   NULL,
							   NULL,
							   NULL);

	DefineCustomIntVariable("telnet_srv.unix_socket_permissions",
							"Telnet server unix socket permissions.",
							NULL,
							&pe_unix_socket_permissions,
							0777,
							0,
							0777,
							PGC_POSTMASTER,
							0,
							NULL,
							NULL,
							NULL);

	DefineCustomStringVariable("telnet_srv.unix_socket_group",
							   "Telnet server unix socket group.",
							   NULL,
							   &pe_unix_socket_group,
							   NULL,
							   PGC_POSTMASTER,
							   0,
							   NULL,
							   NULL,
							   NULL);

	/*
	 * Install hooks.
	 */
	prev_listen_init = listen_init_hook;
	listen_init_hook = pe_listen_init;
}

/*
 * Module unload callback
 */
void
_PG_fini(void)
{
	/* Uninstall hooks. */
	listen_init_hook = prev_listen_init;
}

/*
 * pe_listen_init - Create the telnet server socket(s)
 */
static void
pe_listen_init(void)
{
	pe_create_server_ports();
}

/*
 * pe_accept - Accept a new incoming client connection
 */
static int
pe_accept(pgsocket server_fd, Port *port)
{
	return pe_create_connection(server_fd, port);
}

/*
 * pe_close - called to close server sockets in new backend
 */
static void
pe_close(pgsocket server_fd)
{
	/* nothing to do here */
}

/*
 * pe_start - equivalent of ProcessStartupPacket()
 *
 *	This function needs to communicate with the client
 *	at least to the point where it can fill in the
 *	database_name and user_name in the Port. It cannot
 *	do anything that would require actual database
 *	access.
 */
static int
pe_start(Port *port)
{
	MemoryContext	oldcxt;
	char		   *msg;
	StringInfoData	datname;
	StringInfoData	usename;
	int				i;

	initStringInfo(&datname);
	initStringInfo(&usename);

	msg = "PostgreSQL Telnet Interface\r\n\r\n";
	pq_putbytes(msg, strlen(msg));

	msg = "database name: ";
	pq_putbytes(msg, strlen(msg));
	pq_flush();
	pq_getline(&datname);

	msg = "username: ";
	pq_putbytes(msg, strlen(msg));
	pq_flush();
	pq_getline(&usename);

	/* Copy the values for database and user name into Port */
	oldcxt = MemoryContextSwitchTo(TopMemoryContext);
	port->database_name = pstrdup(datname.data);
	port->user_name = pstrdup(usename.data);
	MemoryContextSwitchTo(oldcxt);

	/* Trim white space at the end of both */
	i = strlen(port->database_name);
	while (i > 0 && isspace(port->database_name[i - 1]))
		port->database_name[--i] = '\0';
	i = strlen(port->user_name);
	while (i > 0 && isspace(port->user_name[i - 1]))
		port->user_name[--i] = '\0';

	resetStringInfo(&datname);
	resetStringInfo(&usename);

	return STATUS_OK;
}

static void
pe_authenticate(Port *port)
{
	char		   *shadow_pass;
	char		   *logdetail = NULL;
	int				result;
	StringInfoData	passwd;
	char		   *msg;
	int				i;

	/*
	 * Perform authentication exchange.
	 */
	set_ps_display("authentication");

	initStringInfo(&passwd);

	/* Get the password */
	msg = "password: ";
	pq_putbytes(msg, strlen(msg));
	pq_flush();
	pq_getline(&passwd);
	i = strlen(passwd.data);
	while (i > 0 && isspace(passwd.data[i - 1]))
		passwd.data[--i] = '\0';

	/* Get the password from pg_authid and verify it */
	shadow_pass = get_role_password(port->user_name, &logdetail);
	if (shadow_pass)
	{
		result = plain_crypt_verify(port->user_name, shadow_pass, passwd.data,
									&logdetail);
	}
	else
		elog(FATAL, "password lookup for user \"%s\" failed", port->user_name);
	if (shadow_pass)
		pfree(shadow_pass);
	resetStringInfo(&passwd);

	if (Log_connections)
	{
		StringInfoData logmsg;

		initStringInfo(&logmsg);
		appendStringInfo(&logmsg, _("connection authorized: user=%s"),
						 port->user_name);

		appendStringInfo(&logmsg, _(" application=telnet"));

		ereport(LOG, errmsg_internal("%s", logmsg.data));
		pfree(logmsg.data);
	}

	set_ps_display("startup");

	ClientAuthInProgress = false;	/* client_min_messages is active now */

	if (result != STATUS_OK)
		elog(FATAL, "authentication failed");
}

static void
pe_mainfunc(Port *port, int argc, char *argv[])
{
	/*
	 * This protocol doesn't need anything other than the default
	 * behavior of PostgresMain(). Note that PostgresMain() will
	 * connect to the database and in turn will call our
	 * pe_authenticate() function.
	 */
	PostgresMain(argc, argv, port->database_name, port->user_name);
}

static void
pe_send_message(ErrorData *edata)
{
	char		   *prefix = "???";
	StringInfoData	msg;

	switch (edata->elevel)
	{
		case DEBUG1:
		case DEBUG2:
		case DEBUG3:
		case DEBUG4:
		case DEBUG5:
			prefix = "DEBUG";
			break;
		case LOG:
		case LOG_SERVER_ONLY:
			prefix = "LOG";
			break;
		case INFO:
			prefix = "INFO";
			break;
		case NOTICE:
			prefix = "NOTICE";
			break;
		case WARNING:
		case WARNING_CLIENT_ONLY:
			prefix = "WARNING";
			break;
		case ERROR:
			prefix = "ERROR";
			break;
		case FATAL:
			prefix = "FATAL";
			break;
		case PANIC:
			prefix = "PANIC";
			break;
	}

	initStringInfo(&msg);
	appendStringInfo(&msg, "Message: %s - %s\r\n", prefix, edata->message);
	pq_putbytes(msg.data, msg.len);
	resetStringInfo(&msg);
}

static void
pe_send_ready_for_query(CommandDest dest)
{
	switch (dest)
	{
		case DestRemote:
		case DestRemoteExecute:
		case DestRemoteSimple:
			pq_putbytes("> ", 2);
			pq_flush();
			break;

		default:
			break;
	}
}

static int
pe_read_command(StringInfo inBuf)
{
	int		rc;

	resetStringInfo(inBuf);
	HOLD_CANCEL_INTERRUPTS();
	pq_startmsgread();
	rc = pq_getline(inBuf);
	if (rc)
		return EOF;

	pq_endmsgread();
	RESUME_CANCEL_INTERRUPTS();

	appendStringInfoChar(inBuf, '\0');
		
	return 'Q';
}

static void
pe_end_command(const QueryCompletion *qc, CommandDest dest, bool force_undecorated_output)
{
	char		completionTag[COMPLETION_TAG_BUFSIZE];
	CommandTag	tag;
	const char *tagname;

	switch (dest)
	{
		case DestRemote:
		case DestRemoteExecute:
		case DestRemoteSimple:

			/*
			 * We assume the tagname is plain ASCII and therefore requires no
			 * encoding conversion.
			 *
			 * We no longer display LastOid, but to preserve the wire
			 * protocol, we write InvalidOid where the LastOid used to be
			 * written.
			 *
			 * All cases where LastOid was written also write nprocessed
			 * count, so just Assert that rather than having an extra test.
			 */
			tag = qc->commandTag;
			tagname = GetCommandTagName(tag);

			if (command_tag_display_rowcount(tag) && !force_undecorated_output)
				snprintf(completionTag, COMPLETION_TAG_BUFSIZE,
						 tag == CMDTAG_INSERT ?
						 "%s 0 " UINT64_FORMAT "\r\n" :
						 "%s " UINT64_FORMAT "\r\n",
						 tagname, qc->nprocessed);
			else
				snprintf(completionTag, COMPLETION_TAG_BUFSIZE, "%s\r\n", tagname);
			pq_putbytes(completionTag, strlen(completionTag));

		case DestNone:
		case DestDebug:
		case DestSPI:
		case DestTuplestore:
		case DestIntoRel:
		case DestCopyOut:
		case DestSQLFunction:
		case DestTransientRel:
		case DestTupleQueue:
			break;
	}
}

/*
 * Get the lookup info that printtup() needs
 */
static void
printtup_prepare_info(DR_printtup *myState, TupleDesc typeinfo, int numAttrs)
{
	int16	   *formats = myState->portal->formats;
	int			i;

	/* get rid of any old data */
	if (myState->myinfo)
		pfree(myState->myinfo);
	myState->myinfo = NULL;

	myState->attrinfo = typeinfo;
	myState->nattrs = numAttrs;
	if (numAttrs <= 0)
		return;

	myState->myinfo = (PrinttupAttrInfo *)
		palloc0(numAttrs * sizeof(PrinttupAttrInfo));

	for (i = 0; i < numAttrs; i++)
	{
		PrinttupAttrInfo *thisState = myState->myinfo + i;
		int16		format = (formats ? formats[i] : 0);
		Form_pg_attribute attr = TupleDescAttr(typeinfo, i);

		thisState->format = format;
		getTypeOutputInfo(attr->atttypid,
						  &thisState->typoutput,
						  &thisState->typisvarlena);
		fmgr_info(thisState->typoutput, &thisState->finfo);
	}
}

static bool
pe_printtup(TupleTableSlot *slot, DestReceiver *self)
{
	TupleDesc	typeinfo = slot->tts_tupleDescriptor;
	DR_printtup *myState = (DR_printtup *) self;
	MemoryContext oldcontext;
	int			natts = typeinfo->natts;
	int			i;

	/* Set or update my derived attribute info, if needed */
	if (myState->attrinfo != typeinfo || myState->nattrs != natts)
		printtup_prepare_info(myState, typeinfo, natts);

	/* Make sure the tuple is fully deconstructed */
	slot_getallattrs(slot);

	/* Switch into per-row context so we can recover memory below */
	oldcontext = MemoryContextSwitchTo(myState->tmpcontext);

	/*
	 * send the attributes of this tuple
	 */
	for (i = 0; i < natts; ++i)
	{
		PrinttupAttrInfo *thisState = myState->myinfo + i;
		Datum		attr = slot->tts_values[i];
		char	   *outputstr;

		if (i > 0)
			pq_putbytes(" | ", 3);

		if (slot->tts_isnull[i])
		{
			pq_putbytes("NULL", 5);
			continue;
		}

		Assert(thisState->format == 0);

		outputstr = OutputFunctionCall(&thisState->finfo, attr);
		pq_putbytes(outputstr, strlen(outputstr));
	}
	pq_putbytes("\r\n", 2);

	/* Return to caller's context, and flush row's temporary memory */
	MemoryContextSwitchTo(oldcontext);
	MemoryContextReset(myState->tmpcontext);

	return true;
}

static void
pe_printtup_startup(DestReceiver *self, int operation, TupleDesc typeinfo)
{
	DR_printtup *myState = (DR_printtup *) self;
	int			i;

	/*
	 * Create I/O buffer to be used for all messages.  This cannot be inside
	 * tmpcontext, since we want to re-use it across rows.
	 */
	initStringInfo(&myState->buf);

	/*
	 * Create a temporary memory context that we can reset once per row to
	 * recover palloc'd memory.  This avoids any problems with leaks inside
	 * datatype output routines, and should be faster than retail pfree's
	 * anyway.
	 */
	myState->tmpcontext = AllocSetContextCreate(CurrentMemoryContext,
												"printtup",
												ALLOCSET_DEFAULT_SIZES);

	/*
	 * If we are supposed to emit row descriptions, then send the tuple
	 * descriptor of the tuples.
	 */
	for (i = 0; i < typeinfo->natts; i++)
	{
		Form_pg_attribute att = TupleDescAttr(typeinfo, i);
		char			 *colname = NameStr(att->attname);

		if (i > 0)
			pq_putbytes(" | ", 3);

		pq_putbytes(colname, strlen(colname));
	}
	pq_putbytes("\r\n----\r\n", 8);

	/* ----------------
	 * We could set up the derived attr info at this time, but we postpone it
	 * until the first call of printtup, for 2 reasons:
	 * 1. We don't waste time (compared to the old way) if there are no
	 *	  tuples at all to output.
	 * 2. Checking in printtup allows us to handle the case that the tuples
	 *	  change type midway through (although this probably can't happen in
	 *	  the current executor).
	 * ----------------
	 */
}

static void
pe_printtup_shutdown(DestReceiver *self)
{
	DR_printtup *myState = (DR_printtup *) self;

	if (myState->myinfo)
		pfree(myState->myinfo);
	myState->myinfo = NULL;

	myState->attrinfo = NULL;

	if (myState->buf.data)
		pfree(myState->buf.data);
	myState->buf.data = NULL;

	if (myState->tmpcontext)
		MemoryContextDelete(myState->tmpcontext);
	myState->tmpcontext = NULL;

	pq_putbytes("\r\n", 2);
}

static void
pe_printtup_destroy(DestReceiver *self)
{
	pfree(self);
}
